import paho.mqtt.client as mqtt
import asyncio
from typing import Dict, Optional, Callable, Any, List
from hwhkit.utils import logger
from hwhkit.connection.mqtt.plugins import PluginBase
from hwhkit.connection.mqtt.plugins.encryption import EncryptionPlugin
from hwhkit.connection.mqtt.plugins.proto import ProtobufPlugin


class MQTTClientManager:
    def __init__(self, add_plugins=True, mqtt_config="mqtt_key_pairs.yaml"):
        self.clients: Dict[str, mqtt.Client] = {}
        self.topic_callbacks: Dict[str, List[Callable[[mqtt.Client, str], Any]]] = {}
        self.plugins = []

        if add_plugins:
            self.plugins.append(EncryptionPlugin(mqtt_config))
            # self.plugins.append(ProtobufPlugin())
            self.reversed_plugins = list(reversed(self.plugins))

    def create_client(
        self,
        client_id: str,
        broker: str,
        port: int = 1883,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> 'MQTTClientManager':
        if client_id in self.clients:
            raise ValueError(f"Client with ID {client_id} already exists.")

        client = mqtt.Client(client_id=client_id)
        if username and password:
            client.username_pw_set(username, password)

        client.broker = broker
        client.port = port
        client.on_connect = self._on_connect
        client.on_disconnect = self._on_disconnect
        client.reconnect_delay_set(min_delay=1, max_delay=120)

        self.clients[client_id] = client
        return self

    def get_client(self, client_id: str) -> Optional[mqtt.Client]:
        return self.clients.get(client_id)

    def remove_client(self, client_id: str):
        client = self.clients.pop(client_id, None)
        if client:
            client.disconnect()

    def add_plugin(self, plugin: PluginBase) -> 'MQTTClientManager':
        self.plugins.append(plugin)
        return self

    def start_all_clients(self) -> 'MQTTClientManager':
        for client_id, client in self.clients.items():
            client.connect(client.broker, client.port, 60)
            client.loop_start()
        return self

    def start_client(self, client_id: str):
        client = self.get_client(client_id)
        if client:
            client.connect(client.broker, client.port, 60)
            client.loop_start()
        else:
            raise ValueError(f"Client with ID {client_id} not found.")

    def _on_connect(self, client: mqtt.Client, userdata, flags, rc):
        if rc == 0:
            logger.info(f"Client {client._client_id.decode()} connected to MQTT Broker!")
        else:
            logger.info(f"Client {client._client_id.decode()} failed to connect, return code {rc}")

    def _on_disconnect(self, client: mqtt.Client, userdata, rc):
        logger.info(f"Client {client._client_id.decode()} disconnected from MQTT Broker!")
        if rc != 0:
            logger.info(f"Unexpected disconnection. Reconnecting...")
            client.reconnect()

    def _on_message(self, topic: str, client: mqtt.Client, msg: mqtt.MQTTMessage):
        logger.info(f"Received topic@{topic}, {msg.payload}")
        try:
            if topic in self.topic_callbacks:
                processed_message = msg.payload
                for plugin in self.reversed_plugins:
                    processed_message = plugin.on_message_received(topic, processed_message)
                for callback in self.topic_callbacks[topic]:
                    callback(client, processed_message)
        except Exception as e:
            logger.error(f"on_message error: {e}")

    def subscribe(self, topic: str):
        def decorator(callback: Callable[[mqtt.Client, str], Any]):
            if topic not in self.topic_callbacks:
                self.topic_callbacks[topic] = []

            self.topic_callbacks[topic].append(callback)

            for client_id, client in self.clients.items():
                client.subscribe(topic)
                client.message_callback_add(
                    topic,
                    lambda c, userdata, msg: self._on_message(topic, c, msg),
                )
            logger.info(f"Subscribed to topic: {topic}")
            return callback

        return decorator

    def publish(self, client_id: str, topic: str, message: str, qos: int = 0, retain: bool = False):
        client = self.get_client(client_id)
        if client:
            if not client.is_connected():
                logger.info(f"Client {client_id} is not connected. Attempting to reconnect...")
                client.reconnect()

            processed_message = message
            for plugin in self.plugins:
                processed_message = plugin.on_message_published(topic, processed_message)

            result = client.publish(topic, processed_message, qos=qos, retain=retain)
            if result.rc == mqtt.MQTT_ERR_SUCCESS:
                logger.info(f"Client {client_id} published message to {topic}")
            else:
                logger.info(f"Client {client_id} failed to publish message to {topic}, error code: {result.rc}")
        else:
            raise ValueError(f"Client with ID {client_id} not found.")


# !/usr/bin/env python
# -*-coding:utf-8 -*-

"""
# Time       ：2025/1/4 11:05
# Author     ：Maxwell
# Description：
"""
import re
import base64
import json
import yaml
from pathlib import Path
from typing import Dict, Optional
from hwhkit.utils.security.rsa import RSACipher
from hwhkit.utils.security.aes import AESCipher
from hwhkit.connection.mqtt.plugins import PluginBase


class EncryptionPlugin(PluginBase):
    def __init__(self, key_pairs_file: str):
        self.topic_keys: Dict[str, Dict[str, str]] = {}
        self.load_key_pairs(key_pairs_file)

    def clean_key(self, key):
        return re.sub(r'\s+', '', key)

    def load_key_pairs(self, key_pairs_file: str):
        if not Path(key_pairs_file).exists():
            raise FileNotFoundError(f"KeyPairs file not found: {key_pairs_file}")

        with open(key_pairs_file, "r") as f:
            data = yaml.safe_load(f)

        if "key_pairs" not in data:
            raise ValueError("Invalid key_pairs file format: missing 'key_pairs' key")

        for topic, keys in self.topic_keys.items():
            # keys['pub_key'] = self.clean_key(keys['pub_key'])
            # keys['pri_key'] = self.clean_key(keys['pri_key'])
            for k in keys:
                if k == 'public':
                    start = "-----BEGIN PUBLIC KEY-----"
                    end = "-----END PUBLIC KEY-----"
                else:
                    start = "-----BEGIN PRIVATE KEY-----"
                    end = "-----END PRIVATE KEY-----"

                if not k.startswith(start) or not k.endswith(end):
                    raise ValueError(f"Invalid {k} key format")

        self.topic_keys = data["key_pairs"]

    def get_keys(self, topic: str) -> Dict[str, str]:
        if topic not in self.topic_keys:
            raise ValueError(f"No keypair found for topic: {topic}")
        return self.topic_keys[topic]

    def on_message_published(self, topic: str, message: str) -> str:
        try:
            keys = self.get_keys(topic)
            public_key_pem = keys["pub_key"]
            aes_cipher = AESCipher()
            aes_key = aes_cipher.get_key()
            public_key_pem_bytes = public_key_pem.encode('utf-8')
            public_key = RSACipher.deserialize_public_key(public_key_pem_bytes)
            rsa_cipher = RSACipher(public_key=public_key)
            encrypted_aes_key = rsa_cipher.encrypt(aes_key.encode('utf-8'))
            encrypted_message = aes_cipher.encrypt(message)

            payload = {
                "k": base64.b64encode(encrypted_aes_key).decode('utf-8'),
                "v": encrypted_message
            }
            return json.dumps(payload)
        except Exception as e:
            raise ValueError(f"Encryption failed: {str(e)}")

    def on_message_received(self, topic: str, message: str) -> str:
        try:
            keys = self.get_keys(topic)
            private_key_pem = keys["pri_key"]

            payload = json.loads(message)
            if 'k' not in payload or 'v' not in payload:
                raise ValueError("Invalid payload format")

            private_key_pem_bytes = private_key_pem.encode('utf-8')
            private_key = RSACipher.deserialize_private_key(private_key_pem_bytes)
            rsa_cipher = RSACipher(private_key=private_key)
            encrypted_aes_key = base64.b64decode(payload['k'].encode('utf-8'))
            aes_key = rsa_cipher.decrypt(encrypted_aes_key).decode('utf-8')

            aes_cipher = AESCipher(key=base64.b64decode(aes_key))
            return aes_cipher.decrypt(payload['v'])
        except Exception as e:
            raise ValueError(f"Decryption failed: {str(e)}")


